/*
 *  paldpl.c: command-line application searching for potential 
 *            palindrome sequences in DNA.
 *
 *  Copyright (C) 2011
 *    Brno University of Technology, Faculty of Information Technology
 *    Masaryk University, Faculty of Informatics
 *
 *  Author(s): Tomas Martinek <martinto@fit.vutbr.cz>
 *             Matej Lexa <lexa@fi.muni.cz> 
 *             
 *  LICENSE TERMS
 *               
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *  3. Neither the name of the Company nor the names of its contributors
 *  may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *                                              
 *  This software is provided ``as is'', and any express or implied
 *  warranties, including, but not limited to, the implied warranties of
 *  merchantability and fitness for a particular purpose are disclaimed.
 *  In no event shall the company or contributors be liable for any
 *  direct, indirect, incidental, special, exemplary, or consequential
 *  damages (including, but not limited to, procurement of substitute
 *  goods or services; loss of use, data, or profits; or business
 *  interruption) however caused and on any theory of liability, whether
 *  in contract, strict liability, or tort (including negligence or
 *  otherwise) arising in any way out of the use of this software, even
 *  if advised of the possibility of such damage.
 */


#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <ctype.h>

/* default parameters */
#define STEPS           10
#define ERR_NUM         4
#define MIN_LEN         4

/* Acceptable command line arguments */
#define ARGUMENTS       "f:s:l:e:vh"

//#define DEBUG
/* --------------------------------------------------------------------------
 *                          Type definition
 * -------------------------------------------------------------------------- 
 */
typedef struct{
        int     diag;
        int     antidiag;
} t_pos;

typedef struct{
        t_pos   start;
        t_pos   lastm;
        int     score;
        int     status;
        int     length;
} t_diag;

/* --------------------------------------------------------------------------
 *                          Procedures and function 
 * -------------------------------------------------------------------------- 
 */

void export_pal(t_diag diag, char ch, char *data)
{
        int     start_ch, end_ch;
        int     start_gap, end_gap;

        /* calculation of string positions */
        end_ch = (diag.lastm.diag + diag.lastm.antidiag - 1)/2;
        start_ch = end_ch - diag.lastm.antidiag;

        end_gap = (diag.start.diag + diag.start.antidiag - 1)/2;
        start_gap = end_gap - diag.start.antidiag;

        printf("%c  % 6d   % 6d   % 6d   % 6d     ", ch,
                start_ch+1, start_gap+1, end_gap+1, end_ch+1);

        for(int i=start_ch; i<=start_gap; i++) 
                printf("%c", data[i]);
        printf("-");
        for(int i=start_gap+1; i<end_gap; i++) 
                printf("%c", data[i]);
        printf("-");
        for(int i=end_gap; i<=end_ch; i++) 
                printf("%c", data[i]);

        printf("\n");
};

void print_score(int* ptr, int size) 
{
        for (int i=0; i<size; i++) 
                printf("%d,", ptr[i]);
        printf("\n");
}

void print_status_array(t_diag* ptr, int size, int border) 
{
        for (int i=0; i<border; i++)
                printf(" ");

        for (int i=border; i<=size-border; i++) { 
                if (ptr[i].status == 0) printf(" ");
                if (ptr[i].status == 1) printf(".");
                if (ptr[i].status == 2) printf("*");
                if (ptr[i].status == 3) printf("|");
        }
        printf("\n");

#ifdef DEBUG
        /* print start */
        printf("    ");
        for (int i=0; i<border; i++) printf(" ");
        for (int i=border; i<=size-border; i++) printf("%d",ptr[i].start.antidiag);
        printf("\n");

        /* print length */
        printf("    ");
        for (int i=0; i<border; i++) printf(" ");
        for (int i=border; i<=size-border; i++) printf("%d",ptr[i].length);
        printf("\n");

        /* print score */
        printf("    ");
        for (int i=0; i<border; i++) printf(" ");
        for (int i=border; i<=size-border; i++) printf("%d",ptr[i].score);
        printf("\n");
#endif

}

int complementary(char a, char b)
{
        return ((a=='a') && (b=='t')) || ((a=='t') && (b=='a')) ||
               ((a=='c') && (b=='g')) || ((a=='g') && (b=='c'));
}

t_diag get_max_score(char a, char b, t_diag dl, t_diag d, t_diag dr, int diag, int antidiag)
{
        t_diag res;

        if (complementary(a,b)) { 
                res = d;
                res.score = d.score+1;
                res.lastm.diag = diag;
                res.lastm.antidiag = antidiag;

                /* the first match */
                if (d.score == 0) {
                        res.start.diag = diag;
                        res.start.antidiag = antidiag;
                }
        }
        else {
                /* mismatch is better */
                if ((d.score-1 >= dl.score-2) && (d.score-1 >= dr.score-2)) {
                        res = d;
                        res.score = d.score-1;
                }
                /* get from left diagonal */
                else if (dl.score > dr.score) {
                        res = dl;
                        res.score = dl.score-2;
                }
                /* get from right diagonal */
                else {
                        res = dr;
                        res.score = dr.score-2;
                }
                /* score underflow */
                if (res.score < 0) {
                        res.score = 0;
                        /* res.status = 0; */
                        /* res.start.diag = diag; */
                        /* res.start.antidiag = antidiag; */
                }
        }

        return res;
}

void usage(char *me)
{
        printf("Usage: %s [-h]\n", me);
        printf("-f s      Input file\n");
        printf("-s num    Number of steps\n");
        printf("-e num    Number of matches per single error\n");
        printf("-l num    Minimal palindrome length\n");
        printf("-v        Verbose output\n");
        printf("-h        Show this text\n");
}



/* --------------------------------------------------------------------------
 *                               Main Program
 * -------------------------------------------------------------------------- 
 */
int main(int argc, char **argv) {
        FILE    *in;
        char    *filename; 
        long    size, read;
        char    *data, *me = argv[0];
        int     steps=STEPS, min_len=MIN_LEN, err_num=ERR_NUM;
        int     x, d, i, num, c, ch, length;
        int     file_flag=0, verbose_flag=0;
        int     exp_num, nexp_num=0;
        t_diag  *diag;
        clock_t c0,c1;


        while ((c = getopt(argc, argv, ARGUMENTS)) != -1) {
                switch (c) {
                case 'f':
                        filename = optarg;
                        file_flag = 1;
                        break;

                case 's':
                        steps = atoi(optarg);
                        break;

                case 'e':
                        err_num = atoi(optarg);
                        break;

                case 'l':
                        min_len = atoi(optarg);
                        break;

                case 'v':
                        verbose_flag = 1;
                        break;

                case 'h':
                        usage(me);
                        return 0;

                default:
                        printf("ERROR: unknown argument -%c\n", optopt);
                	return 1;
                }
        }

        argc -= optind;
        argv += optind;

        if (argc != 0) {
                printf("ERROR: stray arguments\n");
                return 1;
	}

        if (file_flag==0) {
                usage(me);
                return 0;
        }

        /* open input file */
        if ((in = fopen(filename, "rt")) == NULL) {
                printf("ERROR: Can not open file %s.\n", filename);
                return 1;
        }

        /* find size of the file */
        size = 0;
        while((ch = fgetc(in)) != EOF) 
                if (isalpha(ch))
                        size++;
        if (verbose_flag)
                printf("INFO: File size is %ld\n", size);

        /* space alocation and data loading */
        if((data = (char*)malloc(size)) == NULL) {
                printf("ERROR: Not enough space for data[].\n");
                return 1;
        }
        i=0;
        fseek(in, 0, SEEK_SET);
        while((ch = fgetc(in)) != EOF) 
                if (isalpha(ch))
                        data[i++] = ch;
        fclose(in);

        if (verbose_flag)
                printf("INFO: Read data %ld\n", i);

        /* character conversion */
        for (i=0; i<size; i++) 
                data[i] = tolower(data[i]);

        /* print string and its offsets */
        if (verbose_flag) {
                printf("    ");
                for (i=1; i<size; i++) 
                        if (i % 10 == 0)
                                printf("% 20d", i);
                printf("\n");
                printf("Len. ");
                for (i=0; i<size; i++) 
                        printf("%c ", data[i]);
                printf("\n");
        }

        /* diag structure array alocation */
        if((diag = (t_diag*)malloc(2*size*sizeof(t_diag))) == NULL) {
                printf("ERROR: Not enough space for diag[].\n");
                return 1;
        }

        /* find palindromes */
        for (i=0; i<2*size; i++) {
                diag[i].score = 0;
                diag[i].status = 0;
                diag[i].start.diag = i;
                diag[i].start.antidiag = ((i % 2) == 0) ? 1 : 2;
        }

        /* Start point for performance measurement */
        c0 = clock();
        exp_num = 0;

        for (x=1; (x<=2*steps) && (x<size); x++) {
                d = x+1;
                if (verbose_flag & (x % 2 == 0))
                        printf("%02d: ", x/2);

                for (i=x; i<size; i++) {
                        diag[d] = get_max_score(data[i], data[i-x],
                                         diag[d-1], 
                                         diag[d], 
                                         diag[d+1],
                                         d, x);

                        length = (x-diag[d].start.antidiag+1)/2;
                        if (length == 0)
                                length = 1;
                        diag[d].length = length;

                        /* actual score satisfies the required quality (e parameter) */
                        if (length*(err_num-2) <= diag[d].score*err_num) {
                                diag[d].status =  (length >= min_len) ? 3 : 1;

                                /* if palindrome can not continue, then export */
                                if ( ((diag[d].status & 1)==1) && 
                                     ((d==(x+1)) || (d==(2*size-x-1))) &&
                                     (length >= min_len) 
                                   ) {
                                        exp_num++;
                                        diag[d].status = 2;
                                        if (!verbose_flag) {
                                                export_pal(diag[d], 'E', data);
                                        }
                                   }
                        }
                        /* actual score does not satisfies the required quality (e parameter) */
                        else {
                                /* if quality requirement was satisfied in previous step, then export */
                                if(
                                ((diag[d-1].status & 1)==0) &&
                                ((diag[d+1].status & 1)==0) &&
                                ((diag[d].status & 1)==1) &&
                                (length > min_len) ) {
                                        diag[d].status = 2;
                                        exp_num++;
                                        if (!verbose_flag)
                                                export_pal(diag[d], 'E', data);
                                }
                                else 
                                        diag[d].status = 0;
                        }
                        d += 2;
                }

                if (verbose_flag & (x % 2 == 0))
                        print_status_array(diag, 2*size, x);
        }

        /* End point for performance measurement */
        c1 = clock();
        if (verbose_flag)
                printf ("Time:             %f\n", (float)(c1 - c0)/CLOCKS_PER_SEC);

        /* Print nonexported palindromes */
        nexp_num = 0;
        for (i=1; i<(2*size); i++) {
                if ((diag[i].status)==3) {
                        if (!verbose_flag)
                                export_pal(diag[i], 'N', data);
                        nexp_num++;
                }
        }

        /* Print final statistics */
        if (verbose_flag) {
                printf("Exp:              %d\n", exp_num);
                printf("nExp:             %d\n", nexp_num);
                printf("Exp/(Exp+nExp): %8.4f\n", (float)exp_num/(float)(nexp_num+exp_num));
        }

        return 0;
}

